known_fids = {
    "PROD_PERM": ["1", False],
    "DSPLY_NAME": ["3", False],
    "CURRENCY": ["15", True],
    "ACTIV_DATE": ["17", False],
    "RECORDTYPE": ["259", False],
    "RDN_EXCHD2": ["1709", True],
    "QUOTE_DATE": ["3386", False],
    "PROV_SYMB": ["3422", False],
    "PR_RNK_RUL": ["3423", True],
    "OR_RNK_RUL": ["3425", True],
    "MNEMONIC": ["3694", False],
    "QUOTIM_MS": ["3855", False],
    "MKT_STATUS": ["3915", False],
    "TIMACT_MS": ["4148", False],
    "CONTEXT_ID": ["5357", False],
    "DDS_DSO_ID": ["6401", False],
    "SPS_SP_RIC": ["6480", False],
    "BOOK_STATE": ["6516", True],
    "HALT_REASN": ["6517", False],
    "MKT_OR_RUL": ["6519", True],
    "TRD_STATUS": ["6614", True],
    "HALT_RSN": ["6615", True],
    "INST_PHASE": ["8927", True],
    "TIMACT_NS": ["14269", False],
    "BID_TIME": ["266", False],
    "ORDER_PRC": ["3427", False],
    "ORDER_SIDE": ["3428", True],
    "NO_ORD": ["3430", False],
    "ACC_SIZE": ["4356", False],
    "LV_TIM_MS": ["6527", False],
    "LV_DATE": ["6529", False],
    "LV_TIM_NS": ["14268", False],
    "ASK_TIME": ["267", False],
    "ORDBK_VOL": ["3875", False],
}

columns_order = [
    "TICKER",
    "TIMESTAMP",
    "GMT_OFFSET",
    "MARKET_MESSAGE_TYPE",
    "PROD_PERM",
    "DSPLY_NAME",
    "CURRENCY",
    "ACTIV_DATE",
    "RECORDTYPE",
    "RDN_EXCHD2",
    "QUOTE_DATE",
    "PROV_SYMB",
    "PR_RNK_RUL",
    "OR_RNK_RUL",
    "MNEMONIC",
    "QUOTIM_MS",
    "MKT_STATUS",
    "TIMACT_MS",
    "CONTEXT_ID",
    "DDS_DSO_ID",
    "SPS_SP_RIC",
    "BOOK_STATE",
    "HALT_REASN",
    "MKT_OR_RUL",
    "TRD_STATUS",
    "HALT_RSN",
    "INST_PHASE",
    "TIMACT_NS",
    "MAP_ENTRY_TYPE",
    "MAP_ENTRY_KEY",
    "BID_TIME",
    "ASK_TIME",
    "ORDER_PRC",
    "ORDER_SIDE",
    "NO_ORD",
    "ACC_SIZE",
    "LV_TIM_MS",
    "LV_DATE",
    "LV_TIM_NS",
    "ORDBK_VOL",
]

supplement = [
    "TICKER",
    "TIMESTAMP",
    "GMT_OFFSET",
    "MARKET_MESSAGE_TYPE",
    "MAP_ENTRY_TYPE",
    "MAP_ENTRY_KEY",
]
